/* Name: main.c
 * Project: generic clock
 * Author: Leonas (Levas) Vabolis
 * Creation Date: 2009-01-31
 * Copyright: (c) 2009 by Savel (aka Levas). http://www.vabolis.lt , http://www.savel.org
 * License: GNU GPL v2, check includes for more (c) stuff.
 * Hardware: SAVEL ATMEGA16 testing board. Intersil ISL1208 RTC chip. LCD. Analogue buttons.
 */

#define 	F_CPU 12000000	
#include	<avr\io.h>	
#include <stdlib.h>	
#include <avr/interrupt.h>
#include <util\delay.h>
#include <avr/pgmspace.h>
#include "lcd.c"
#include "i2c.c"
#include <string.h>

//Set I2C address of RTC device. I am using Intersil isl1208 chip.
#define RTC_ADDR 0x6F // 1101 111x

// RTC
#define RTC_SC 0x00
#define RTC_MN 0x01
#define RTC_HR 0x02
#define RTC_DT 0x03
#define RTC_MO 0x04
#define RTC_YR 0x05
#define RTC_DW 0x06

//CNTR & STATUS
#define RTC_SR 0x07
#define RTC_INT 0x08
#define RTC_ATR 0x0A
#define RTC_DTR 0x0B

//ALARMS
#define RTC_SCA 0x0C
#define RTC_MNA 0x0D
#define RTC_HRA 0x0E
#define RTC_DTA 0x0F
#define RTC_MOA 0x10
#define RTC_DWA 0x11

//Two user bytes
#define RTC_USR1 0x12
#define RTC_USR2 0x13

////Clock config
#define MAX_MENIU 5
#define AUTOKILL 60

#define DEBUG	0



unsigned char hour,minute, sec, dow, day, month, year, ticks,button, bdown, user1, user2;
unsigned char ahour, aminute, asec;
volatile unsigned char status, meniu, display,kill,substatus;
volatile char value;
volatile char hour_mil=128;

int abutton; //analog value of pressed button: 0-1024.
char buffer[7]; //numbers2text buffer.
int x;
unsigned char dst[2]; //I2C stuff buffer.


void print_hex(unsigned char a)
{
	itoa( a, buffer, 16);
	lcd_puts(buffer);
}
void print_dec(unsigned char a)
{
	itoa( a, buffer, 10);
	lcd_puts(buffer);
}
void print(char *pointer)
{
lcd_puts(pointer);
}
//


int RTC_writeRegister(char i2c_addr, char reg_addr, unsigned char *dat, char len) //Write data to I2C
{
	unsigned char tmp[len+1];
	tmp[0] = reg_addr & 0x1F;
	memcpy(&tmp[1], dat, len);
	return i2c_transaction(i2c_addr, len + 1, tmp, 0, NULL);
}
int RTC_readRegister(char i2c_addr, char reg_addr, unsigned char *dst, char len) //Read data from I2C
{
	unsigned char tmp;
	tmp = reg_addr & 0x1F;
	return i2c_transaction(i2c_addr, 1, &tmp, len, dst);
}

int RTC_configure(char i2c_addr, unsigned char cfg) //Write byte to config register.
{
	return RTC_writeRegister(i2c_addr, RTC_SR, &cfg, 1);
}


void error(void) //Damn, error. You will not see this in normal clock. But if you have problems with I2C...
{

lcd_puts("I2C error");
_delay_ms(2);
i2c_init();
_delay_ms(20);

}

// biski BCD<-->DEC matematikos
// Binary decimal<-->decimal mathematics.
unsigned char bcd(unsigned char dec)
{ return ((dec/10)<<4)+(dec%10); }
unsigned char decimal(unsigned char bcd)
{ return ((bcd>>4)*10)+bcd%16; }



void initAll(void)				
{
//Some hardware inits. (timers, ADC, I2C and set RTC to writable mode).

	//NAUJAS, Timeris 0.
	TCCR0=_BV(CS00)|_BV(CS02);
	
    /* Enable timer 0 overflow interrupt. */
    TIMSK = _BV (TOIE0);

	// ADC init
	ADMUX = _BV(REFS0); // ADC Vref=AVCC, ADC0=in
	ADCSRA |=_BV(ADEN) | _BV(ADPS0) | _BV(ADPS1) | _BV(ADPS2); // enable, ADC prescaler 128
	
	i2c_init();
	RTC_configure(RTC_ADDR, 0x90); //auto reset, write enable
}
void writedate(unsigned char a, unsigned char b, unsigned char c, unsigned char d)
{
// y,m,d,dow
unsigned char i;
	i=bcd(a);
	RTC_writeRegister(RTC_ADDR, RTC_YR, &i, 1);
	_delay_ms(4);
	i=bcd(b);
	RTC_writeRegister(RTC_ADDR, RTC_MO, &i, 1);
	_delay_ms(4);
	i=bcd(c);
	RTC_writeRegister(RTC_ADDR, RTC_DT, &i, 1);
	_delay_ms(4);
	i=bcd(d);
	RTC_writeRegister(RTC_ADDR, RTC_DW, &i, 1);

}
void writehours(unsigned char a, unsigned char b, unsigned char c) // Write new hours to RTC.
{
unsigned char i;
	i=bcd(c);
	RTC_writeRegister(RTC_ADDR, RTC_SC, &i, 1);
	_delay_ms(4);
	i=bcd(b);
	RTC_writeRegister(RTC_ADDR, RTC_MN, &i, 1);
	_delay_ms(4);
	i=bcd(a) | hour_mil;
	RTC_writeRegister(RTC_ADDR, RTC_HR, &i, 1);
}

void  clearscreen(void) //Just clear screen. For hardware independant clock.
{
lcd_clrscr();
}

unsigned char kalendorius(unsigned char year, unsigned char month, unsigned char day)
{

unsigned char a=(14 - month)/12;
unsigned int y=2000+year-a;
unsigned char m=month + 12*a-2;
return (7000+(day+y+y/4-y/100+y/400+(31*m)/12)) % 7;

}
unsigned char leapyear(unsigned char year)
{
int y=year+2000;
unsigned char a=0;

if((y % 4)==0) {a=1;}
if((y % 100)==0) {
	if((y % 400) != 0) {a=0;}
	}
return a;
}

#if DEBUG
void lcddebug(void) //My debug
{
//DEBUG
	lcd_gotoxy(3,1);
	lcd_puts("M:");
	print_hex(meniu);
	
	lcd_puts("S:");
	print_hex(status);

	lcd_puts("B:");
	print_hex(button);

	lcd_puts("K:");
	print_hex(substatus);

}
#endif
void showtime(void) //Direct print from RTC. (Note BCD data)
{
int x;
	//lcd_gotoxy(0,0);
	lcd_home();

x = RTC_readRegister(RTC_ADDR, RTC_HR, &dst[0], 1);
			if (x==0){
			x=dst[0] & 127;

			if(x<10){lcd_puts("0");} //nelabai svarbu, tinka ir 10. HEX:01...09,10,11... DEC: 01...09,16,17...
			print_hex(x);

			} else {error(); }
x = RTC_readRegister(RTC_ADDR, RTC_MN, &dst[0], 1);
			if (x==0){
			x=dst[0];
			
			lcd_puts(":");
			if(x<10){lcd_puts("0");}
			print_hex(x);
			} else {error(); }
x = RTC_readRegister(RTC_ADDR, RTC_SC, &dst[0], 1);
			if (x==0){
			x=dst[0];
			
			lcd_puts(":");
			if(x<10){lcd_puts("0");}
			print_hex(x);
			} else {error();}
			lcd_puts("  ");

}
void showdate(void) //Direct print from RTC. (Note BCD data)
{
int x;
//lcd_gotoxy(0,0);			
lcd_home();
x = RTC_readRegister(RTC_ADDR, RTC_YR, &dst[0], 1);
			if (x==0){
			x=dst[0];
			lcd_puts("20");
			if(x<10) lcd_puts("0");
			print_hex(x);
			} else {error();}
x = RTC_readRegister(RTC_ADDR, RTC_MO, &dst[0], 1);
			if (x==0){
			x=dst[0];
			lcd_puts(",");
			print_hex(x);
			} else {error();}

x = RTC_readRegister(RTC_ADDR, RTC_DT, &dst[0], 1);
			if (x==0){
			x=dst[0];
			lcd_puts(",");
			print_hex(x);
			} else {error();}
x = RTC_readRegister(RTC_ADDR, RTC_DW, &dst[0], 1);
			if (x==0){
			x=dst[0];
			lcd_puts(":");
			print_hex(x);
			} else {error();}
			
			lcd_puts("  ");

}

void readdatafromRTC(void) //Yeap, read data from RTC chip. Convert BCD to DEC.
{
int x;
x = RTC_readRegister(RTC_ADDR, RTC_HR, &dst[0], 1);
			if (x==0){
			hour=decimal(dst[0] & 127);
			hour_mil=dst[0]&128;
			} else {error();}
_delay_ms(4);
x = RTC_readRegister(RTC_ADDR, RTC_MN, &dst[0], 1);
			if (x==0){
			minute=decimal(dst[0]);
			} else {error();}
_delay_ms(4);
x = RTC_readRegister(RTC_ADDR, RTC_SC, &dst[0], 1);
			if (x==0){
			sec=decimal(dst[0]);
			} else {error();}
_delay_ms(4);
x = RTC_readRegister(RTC_ADDR, RTC_DW, &dst[0], 1);
			if (x==0){
			dow=dst[0];
			} else {error();}
_delay_ms(4);
x = RTC_readRegister(RTC_ADDR, RTC_DT, &dst[0], 1);
			if (x==0){
			day=decimal(dst[0]);
			} else {error();}
_delay_ms(4);
x = RTC_readRegister(RTC_ADDR, RTC_MO, &dst[0], 1);
			if (x==0){
			month=decimal(dst[0]);
			} else {error();}
_delay_ms(4);
x = RTC_readRegister(RTC_ADDR, RTC_YR, &dst[0], 1);
			if (x==0){
			year=decimal(dst[0]);
			} else {error();}
_delay_ms(4);			
			

x = RTC_readRegister(RTC_ADDR, RTC_USR1, &dst[0], 1);
			if (x==0){
			user1=dst[0];
			} else {error();}
_delay_ms(4);
x = RTC_readRegister(RTC_ADDR, RTC_USR2, &dst[0], 1);
			if (x==0){
			user2=dst[0];
			} else {error();}
_delay_ms(4);	



x = RTC_readRegister(RTC_ADDR, RTC_SCA, &dst[0], 1);
			if (x==0){
			asec=decimal(dst[0]);
			} else {error();}
_delay_ms(4);	
x = RTC_readRegister(RTC_ADDR, RTC_MNA, &dst[0], 1);
			if (x==0){
			aminute=decimal(dst[0]);
			} else {error();}
_delay_ms(4);	
x = RTC_readRegister(RTC_ADDR, RTC_HRA, &dst[0], 1);
			if (x==0){
			ahour=decimal(dst[0]);
			} else {error();}
	











}
void printset(unsigned char a, unsigned char b, unsigned char c,unsigned char d) // print A,B,C value for setup, emphasize number D
{
			lcd_clrscr();
			//lcd_gotoxy(0,0);
			lcd_home();
			if(a<10) {lcd_puts("0");}
			print_dec(a);
			lcd_puts(":");
			if(b<10) {lcd_puts("0");}
			print_dec(b);
			lcd_puts(":");
			if(c<10) {lcd_puts("0");}	
			print_dec(c);
			lcd_gotoxy(d*3,1);
			lcd_puts("^");

}
void printset2(unsigned char a, unsigned char b, unsigned char c, unsigned char d, unsigned char e)
{
	lcd_clrscr();
	//lcd_gotoxy(0,0);
	lcd_home();
	lcd_puts("20");
	if(a<10) {lcd_puts("0");}			
	print_dec(a);
	lcd_gotoxy(4,0);
	lcd_puts(":");
	print_dec(b);
	lcd_gotoxy(7,0);
	lcd_puts(":");
	print_dec(c);
	lcd_gotoxy(10,0);
	lcd_puts(",");
	print_dec(d);
	lcd_puts(" ");
	lcd_gotoxy(e*3+2,1);
	lcd_puts("^");
}
void printmeniu(unsigned char i) //print Meniu text according menu number.
{
	//lcd_gotoxy(0,0);
	lcd_home();

switch(i)
	{
	case 0:lcd_puts("SET TIME"); break;
	case 1:lcd_puts("SET DATE "); break;
	case 2:lcd_puts("SET ALARM"); break;
	case 3:lcd_puts("SETUP    "); break;
	case 4:lcd_puts("EXIT "); break;
	case 5:lcd_puts("ABOUT   "); break;
	case 6:lcd_puts("entered "); break;
	
	case 7 :lcd_puts("AM/PM-24H:"); break;
	case 8 :lcd_puts("Day dim:  "); break;
	case 9 :lcd_puts("Night dim:"); break;
	case 10:lcd_puts("Day time:    "); break;
	case 11:lcd_puts("Factory reset"); break;
	}
}
void printabout(void) //very nice about text. :)
{
lcd_clrscr();
lcd_puts("SAVEL CLOCK\n");
lcd_puts("V1.0 (c)2009");
}

//---pertraukimas pagal taimeri 0----
ISR(TIMER0_OVF_vect) //timer overflow vector
{
cli();
//This is loooong routine. We do not want retriger to it back. Disable interrupts. (I think is not very usefull as MCU is fast enough to run it)

ADCSRA |=_BV(ADSC); // START ADC READ
ticks++; //magic slowdown counter. Check it in "slow interrupt" part.


//FAST PART OF INTERRUPT**************************



//MYGTUKAI - BUTTONS is everything. We have six of them. And may add more, using only ONE pin of MCU! 'cause we use analog ones.

loop_until_bit_is_set(ADCSRA, ADIF); //uzsiciklinam kol nebaigs darba ADC. (forced delay for ADC. If you have something to do before checking buttons, place it before this line.

abutton=ADCW;
// Analogue buttons, set constants according resistor divider values.
if(abutton<50)
	{button=1; }
else if (abutton<700L)
	{button=2; }
else if (abutton<900L)
	{button=3; }
else if (abutton<960L)
	{button=4; }
else if (abutton<980L)
	{button=5; }
else if (abutton<1009L)
	{button=6; }
else
	{button=0; bdown=0;} //All buttons off. Important!



//MEGA MAGIC begin here. :)
	
if(bdown==0){
// FIRST MENIU LEVEL (TOP LEVEL, NORMAL STATE)***************************
	if(status==0)
	{
	switch(button)
		{
			case 5: status=1; kill=AUTOKILL; bdown=1; break;
			case 4: if(display==0){display=1; clearscreen(); showdate();} else {display=0; clearscreen(); showtime();} break;
			case 1: display=0; showtime(); break;
			case 6: clearscreen(); break;
		}
	}
	if(status==1 &&  bdown==0)
// SECOND MENIU LEVEL
	{
	if(button != 0) {kill=AUTOKILL;}
	switch(button)
		{
			case 6: status=0; clearscreen(); break;
			case 1: meniu--; break;
			case 2: meniu++; break;
			case 5: status=2; bdown=1; 
					readdatafromRTC();
					clearscreen();
					break;
		}
		
		if(meniu==255) {meniu=MAX_MENIU;}
		if(meniu>MAX_MENIU) {meniu=0;} //roll over meniu
		if(status==1){
		printmeniu(meniu);
		}
		
	}
//***********END OF MENIU LEVEL 2 (second)

// THIRD MENIU LEVEL (3)*****************************
	if( bdown==0){ //bdown2
	
	if(status==2 && meniu==5) 
//SELECTED ABOUT
	{
	printabout();
	switch(button)
		{
			case 6: status=0; clearscreen(); break;
			case 5: status=0; clearscreen(); break;
		}
		
	}
	
	if(status==2 && meniu==4) // EXIT
//SELECTED EXIT MENIU. (For those, who didn't have nice button like this:[X] (exit))
	{ status=0; meniu=0; clearscreen(); }


//SELECTED SET DATE
	if(status==2 && meniu==1)
	{
	if(button != 0) {kill=AUTOKILL;}
	
	switch(button)
		{
			case 6: status=1; clearscreen(); value=0; printmeniu(meniu); break;
			case 1: if(substatus<1) {substatus=3;}
					substatus--; value=0; break;
			case 2: substatus++; value=0; break;
			case 4: value=1; break;
			case 3: value=-1; break;
			case 5: writedate(year, month, day, dow); value=0; status=1; clearscreen(); break;
		}
		if (substatus>2) {substatus=0;}
		if(status==2){
		switch(substatus)
		{
			case 0: year=year+value;
					if(year==255) {year=99;}
					if(year>99) {year=0;}
					break;
			case 1: month=month+value;
					if(month==0) {month=12;}
					if(month>12) {month=1;}
					break;
			case 2: day=day+value;
					if(day==0) {day=31;}
					if(day>31) {day=1;}

					break;
		}
		
		//SUPER Calendar- checking for leap year, number of days in a month and calculating weekday. (in 21 century only) :), also proper rollover
		if((day>30) && ((month==4) | (month==6) | (month==9) | (month==11))){day=30; if(value==1){day=1;}}
		if(month==2){if(leapyear(year)==0 && day>28) {day=28; if(value==1){day=1;}} else if (leapyear(year)==1 && day>29) {day=29; if(value==1){day=1;}} }
		dow=kalendorius(year, month,day);
		

		printset2(year, month, day, dow, substatus);
		}
		value=0;
	}
	
//SELECTED SET ALARM
	if(status==2 && meniu==2)
	{
//Cool hardware alarm. Can be programmed to trigger every day, or evry hour, etc. Like cron daemon...
	if(button != 0) {kill=AUTOKILL;}
	switch(button)
		{
			case 6: status=1; clearscreen(); printmeniu(meniu); value=0; break;
			case 1: 
					if(substatus<1) {substatus=3;}
					substatus--; value=0; break; 
			case 2: substatus++; value=0; break; 
			case 4: value=1; break;
			case 3: value=-1; break;
			case 5: value=0;
					status=1;
					clearscreen();
					break; ///<---write to RTC
		}
	if (substatus>2) {substatus=0;}
	if (status==2) {
	switch(substatus)
		{
			case 0: ahour=ahour+value;
					if(ahour==255) {ahour=23;}
					if(ahour>23) {ahour=0;}
					break;
			case 1: aminute=aminute+value;
					if(aminute==255) {aminute=59;}
					if(aminute>59) {aminute=0;}
					break;
			case 2: asec=asec+value;
					if(asec==255) {asec=59;}
					if(asec>59) {asec=0;}
					break;
		}
		printset(ahour,aminute,asec,substatus);}
		value=0;

	}
//SELECTED SETUP
	if(status==2 && meniu==3)
	{
	if(button != 0) {kill=AUTOKILL;}	
	switch(button)
		{
			case 6: status=1; clearscreen(); value=0; printmeniu(meniu); break;
			case 1: if(substatus<1) {substatus=5;}
					substatus--; value=0; break;
			case 2: substatus++; value=0; break;
			case 4: value=1; break;
			case 3: value=-1; break; 
		}
		if (substatus>4) {substatus=0;}
		if(status==2){
		printmeniu(substatus+7);
		switch(substatus)
		{
			case 0: //AM/PM bitas.
					if(value != 0) {hour_mil ^=_BV(7);} 
					if(hour_mil==128) print("MIL  "); else print("AM/PM");
					break;
			case 1: //user1
					user1=user1+value; print_hex(user1); break;
			case 2: //user2
					user2=user2+value; print_hex(user2); break;
					
		}
		}
		value=0;

	}
//SELECTED SET TIME

	if(status==2 && meniu==0)
	{
	if(button != 0) {kill=AUTOKILL;}
	switch(button)
		{
			case 6: status=1; printmeniu(meniu); value=0; break;
			case 1: 
					if(substatus<1) {substatus=3;}
					substatus--; value=0; break; 
			case 2: substatus++; value=0; break; 
			case 4: value=1; break;
			case 3: value=-1; break;
			case 5: writehours(hour, minute, sec); value=0; status=1; break;
		}
	if (substatus>2) {substatus=0;}
	if(status==2) {
	switch(substatus)
		{
			case 0: hour=hour+value;
					if(hour==255) {hour=23;}
					if(hour>23) {hour=0;}
					break;
			case 1: minute=minute+value;
					if(minute==255) {minute=59;}
					if(minute>59) {minute=0;}
					break;
			case 2: sec=sec+value;
					if(sec==255) {sec=59;}
					if(sec>59) {sec=0;}
					break;
		}
		printset(hour,minute,sec,substatus);
		}
		value=0;
	}

	}//bdown2


	if(button != 0) bdown=1;
}//bdown1
//******************END OF FAST INTERUPT PART


// SLOW PART OF INTERRUPT
if(ticks>9){
#if DEBUG
	lcddebug();
#endif
//ANY LEVEL MENIU AUTOKILL. Just leave buttons untouched for some time and device resets to default run level 0 	
	if(kill>1) {kill--;} else {status=0; clearscreen();}


//RUN LEVEL 0 display: default screen of device with outo uptdate.
	if(status==0){
		if (display==0) showtime();
		if (display==1) showdate();
		}
	ticks=0;
	}
	sei();
//It's done. Now we enable interrups for next run.	
}



int main(void)
{

      // Initialize Hardware

	lcd_init(LCD_DISP_ON);
	clearscreen();
	printabout();
	initAll();
	_delay_ms(500);
    // enable all interrupts
    sei();
	  
    for (;;) { 
   
   //Do something intensive here. Like dynamic video display.
   
   
		}
    return 0;
}